@@ -9,24 +9,30 @@ use ExtUtils::testlib;
 use Getopt::Long;
 use Test::More;
 use Data::Dumper;
+use File::Temp qw(tempfile);
 
 use base qw(Exporter);
-our @EXPORT = qw(run_test_group);
+our @EXPORT = qw(
+    run_test_group
+    run_command
+    run_perl_command
+);
 
+use Devel::NYTProf::Data;
 use Devel::NYTProf::Reader;
 use Devel::NYTProf::Util qw(strip_prefix_from_paths html_safe_filename);
 
 
-my $profile_datafile = 'nytprof_t.out';     # non-default to test override works
-
-my $tests_per_extn = {p => 1, rdt => 1, x => 3};
-
+my $this_perl = $^X;
+$this_perl .= $Config{_exe} if $^O ne 'VMS' and $this_perl !~ m/$Config{_exe}$/i;
 
 my %opts = (
-    profperlopts => '-d:NYTProf',
+    one          => $ENV{NYTPROF_TEST_ONE},
+    profperlopts => $ENV{NYTPROF_TEST_PROFPERLOPTS} || '-d:NYTProf',
     html         => $ENV{NYTPROF_TEST_HTML},
+    mergerdt     => $ENV{NYTPROF_TEST_MERGERDT}, # overkill, but handy
 );
-GetOptions(\%opts, qw/p=s I=s v|verbose d|debug html open profperlopts=s leave=i use_db_sub=i savesrc=i compress=i abort/)
+GetOptions(\%opts, qw/p=s I=s v|verbose d|debug html open profperlopts=s leave=i use_db_sub=i savesrc=i compress=i one abort/)
     or exit 1;
 
 $opts{v}    ||= $opts{d};
@@ -52,8 +58,11 @@ for my $opt (qw(trace)) {
 }
 
 
-# but we'll force a specific test data file
-$NYTPROF_TEST{file} = $profile_datafile;
+my $tests_per_extn = {
+    p   => 1,
+    rdt => ($opts{mergerdt}) ? 2 : 1,
+    x   => 3
+};
 
 chdir('t') if -d 't';
 
@@ -63,6 +72,7 @@ if (-d '../blib') {
 my $bindir      = (grep {-d} qw(./blib/script ../blib/script))[0];
 my $nytprofcsv  = "$bindir/nytprofcsv";
 my $nytprofhtml = "$bindir/nytprofhtml";
+my $nytprofmerge= "$bindir/nytprofmerge";
 
 my $path_sep = $Config{path_sep} || ':';
 my $perl5lib = $opts{I} || join($path_sep, @INC);
@@ -71,93 +81,90 @@ my $perl     = $opts{p} || $^X;
 # turn ./perl into ../perl, because of chdir(t) above.
 $perl = ".$perl" if $perl =~ m|^\./|;
 
-my @test_opt_leave      = (defined $opts{leave})      ? ($opts{leave})      : (1, 0);
+
+if ($opts{one}) {           # for one quick test
+    $opts{leave}      = 1;
+    $opts{use_db_sub} = 0;
+    $opts{savesrc}    = 1;
+    $opts{compress}   = 1;
+}
+
+# force savesrc off for perl 5.11.2 due to perl bug RT#70804
+$opts{savesrc} = 0 if $] eq "5.011002";
+
+my @test_opt_leave      = (defined $opts{leave})      ? ($opts{leave})      : (0, 1);
 my @test_opt_use_db_sub = (defined $opts{use_db_sub}) ? ($opts{use_db_sub}) : (0, 1);
 my @test_opt_savesrc    = (defined $opts{savesrc})    ? ($opts{savesrc})    : (0, 1);
 my @test_opt_compress   = (defined $opts{compress})   ? ($opts{compress})   : (0, 1);
 
-# build @env_combinations
-my @env_combinations;
-for my $leave (@test_opt_leave) {
-    for my $use_db_sub (@test_opt_use_db_sub) {
-        for my $savesrc (@test_opt_savesrc) {
-            for my $compress (@test_opt_compress) {
-                push @env_combinations, {
-                    start      => 'init',
-                    leave      => $leave,
-                    use_db_sub => $use_db_sub,
-                    savesrc    => $savesrc,
-                    compress   => $compress,
+sub mk_opt_combinations {
+    my ($overrides) = @_;
+
+    my @opt_combinations;
+    my %seen;
+    for my $leave (@test_opt_leave) {
+        for my $use_db_sub (@test_opt_use_db_sub) {
+            for my $savesrc (@test_opt_savesrc) {
+                for my $compress (@test_opt_compress) {
+                    my $o = {
+                        start      => 'init',
+                        slowops    => 2,
+                        leave      => $leave,
+                        use_db_sub => $use_db_sub,
+                        savesrc    => $savesrc,
+                        compress   => $compress,
+                        ($overrides) ? %$overrides : (),
+                    };
+                    my $key = join "\t", map { "$_=>$o->{$_}" } sort keys %$o;
+                    next if $seen{$key}++;
+                    push @opt_combinations, $o;
                 }
             }
         }
     }
+    return \@opt_combinations;
 }
 
+my %env_influence;
 
-sub run_test_group {
-    my ($opts) = @_;
-    my $override_env     = $opts->{override_env} || {};
-    my $extra_test_code  = $opts->{extra_test_code};
-    my $extra_test_count = $opts->{extra_test_count} || 0;
 
-    # obtain group from file name
-    my $group;
-    if ((caller)[1] =~ /([^\/\\]+)\.t$/) {
-        $group = $1;
-    } else {
-        croak "Can't determine test group";
-    }
+sub do_foreach_opt_combination {
+    my ($opt_combinations, $code) = @_;
 
-    my @tests = grep { -f $_ } map { join('.', $group, $_) } sort keys %$tests_per_extn;
+    COMBINATION:
+    for my $env (@$opt_combinations) {
 
-    if ($opts{v}) {
-        print "tests: @tests\n";
-        print "perl: $perl\n";
-        print "perl5lib: $perl5lib\n";
-        print "nytprofcvs: $nytprofcsv\n";
-    }
-
-    my $tests_per_env = number_of_tests(@tests) + $extra_test_count;
-
-    plan tests => 1 + $tests_per_env * @env_combinations;
-
-    # Windows emulates the executable bit based on file extension only
-    ok($^O eq "MSWin32" ? -f $nytprofcsv : -x $nytprofcsv, "Found nytprofcsv as $nytprofcsv");
-
-    my %env_influence;
-
-    for my $env (@env_combinations) {
         my $prev_failures = count_of_failed_tests();
 
-        my %env = (%$env, %$override_env, %NYTPROF_TEST);
-        local $ENV{NYTPROF} = join ":", map {"$_=$env{$_}"} sort keys %env;
-        my $context = "NYTPROF=$ENV{NYTPROF}\n";
-        ($opts{v}) ? warn $context : print $context;
+        my %env = (%$env, %NYTPROF_TEST);
+        my @keys = sort keys %env; # put trace option first:
+        @keys = ('trace', grep { $_ ne 'trace' } @keys) if $env{trace};
 
-        for my $test (@tests) {
-            run_test($test);
-        }
+        local $ENV{NYTPROF} = join ":", map {"$_=$env{$_}"} @keys;
 
-        if ($extra_test_code) {
-            note("running $extra_test_count extra tests...");
-            my $profile = eval { Devel::NYTProf::Data->new({filename => $profile_datafile}) };
-            if ($@) {
-                diag($@);
-                fail("extra tests group '$group'") foreach (1 .. $extra_test_count);
-                return;
-            }
+        my $context_msg = "NYTPROF=$ENV{NYTPROF}\n";
+        ($opts{v}) ? warn $context_msg : print $context_msg;
 
-            $extra_test_code->($profile, $env);
+        ok eval { $code->(\%env) };
+        if ($@) {
+            diag "Test group aborted: $@";
+            last COMBINATION;
         }
 
         # did any tests fail?
         my $failed = (count_of_failed_tests() - $prev_failures) ? 1 : 0;
         # record what env settings may have influenced the failure
-        ++$env_influence{$_}{$env->{$_}}{$failed ? 'fail' : 'pass'} for keys %$env;
+        ++$env_influence{$_}{$env->{$_}}{$failed ? 'fail' : 'pass'}
+            for keys %$env;
     }
+}
+
+
+# report which env vars influenced the failures, if any
+sub report_env_influence {
+    my ($tag) = @_;
+    #warn Dumper(\%env_influence);
 
-    # report which env vars influenced the failures, if any
     my @env_influence;
     for my $envvar (sort keys %env_influence) {
         my $variants = $env_influence{$envvar};
@@ -173,16 +180,89 @@ sub run_test_group {
         push @env_influence, sprintf "%15s: %s\n", $envvar,
             join ', ', map { "$_ => $variants->{$_}" } sort keys %$variants;
     }
-    if (@env_influence) {
-        diag "Test failures of $group related to settings:";
+    %env_influence = ();
+
+    if (@env_influence and not defined wantarray) {
+        diag "SUMMARY: Breakdown of $tag test failures by option settings:";
         diag $_ for @env_influence;
     }
 
+    return @env_influence;
+}
+
+
+# execute a group of tests (t/testFoo.*) - calls plan()
+sub run_test_group {
+    my ($rtg_opts) = @_;
+    my $extra_test_code  = $rtg_opts->{extra_test_code};
+    my $extra_test_count = $rtg_opts->{extra_test_count} || 0;
+    my $extra_options    = $rtg_opts->{extra_options};
+
+    # obtain group from file name
+    my $group;
+    if ((caller)[1] =~ /([^\/\\]+)\.t$/) {
+        $group = $1;
+    } else {
+        croak "Can't determine test group";
+    }
+
+    my @tests = grep { -f $_ } map { "$group.$_" } sort keys %$tests_per_extn;
+    unlink <$group.*_new{,p}>; # delete _new and _newp files from previous run
+
+    if ($opts{v}) {
+        print "tests: @tests\n";
+        print "perl: $perl\n";
+        print "perl5lib: $perl5lib\n";
+        print "nytprofcvs: $nytprofcsv\n";
+    }
+
+    plan skip_all => "No '$group.*' test files and no extra_test_code"
+        if !@tests and !$extra_test_code;
+
+    my $opts = mk_opt_combinations($extra_options);
+    my $tests_per_env = number_of_tests(@tests) + $extra_test_count + 1;
+
+    plan tests => 1 + $tests_per_env * @$opts;
+
+    # Windows emulates the executable bit based on file extension only
+    ok($^O eq "MSWin32" ? -f $nytprofcsv : -x $nytprofcsv, "Found nytprofcsv as $nytprofcsv");
+
+    # non-default output file to test override works and to allow parallel testing
+    my $profile_datafile = "nytprof_$group.out";
+    $NYTPROF_TEST{file} = $profile_datafile;
+
+    do_foreach_opt_combination( $opts, sub {
+        my ($env) = @_;
+
+        for my $test (@tests) {
+            run_test($test, $env);
+        }
+
+        if ($extra_test_code) {
+            my $profile;
+            if (@tests) {
+                print("running $extra_test_count extra tests...\n");
+                $profile = eval { Devel::NYTProf::Data->new({ filename => $profile_datafile }) };
+                if ($@) {
+                    diag($@);
+                    fail("extra tests group '$group'") foreach (1 .. $extra_test_count);
+                    return;
+                }
+            }
+
+            $extra_test_code->($profile, $env);
+        }
+
+        return 1;
+    } );
+
+    report_env_influence($group);
 }
 
 
 sub run_test {
-    my ($test) = @_;
+    my ($test, $env) = @_;
+    my $tag = join " ", map { ($_ ne 'file') ? "$_=$env->{$_}" : () } sort keys %$env;
 
     #print $test . '.'x (20 - length $test);
     $test =~ / (.+?) \. (?:(\d)\.)? (\w+) $/x or do {
@@ -190,29 +270,43 @@ sub run_test {
         return;
     };
     my ($basename, $fork_seqn, $type) = ($1, $2 || 0, $3);
+    #warn "($basename, $fork_seqn, $type)\n";
 
+    my $profile_datafile = $NYTPROF_TEST{file};
     my $test_datafile = (profile_datafiles($profile_datafile))[$fork_seqn];
+    my $outdir = $basename.'_outdir';
 
     if ($type eq 'p') {
         unlink_old_profile_datafiles($profile_datafile);
-        profile($test, $profile_datafile);
-    }
-    elsif ($type eq 'rdt') {
-        verify_data($test, $test_datafile);
-    }
-    elsif ($type eq 'x') {
-        my $outdir = $basename.'_outdir';
-        mkdir $outdir or die "mkdir($outdir): $!" unless -d $outdir;
-        unlink <$outdir/*>;
-
-        verify_csv_report($test, $test_datafile, $outdir);
+        profile($test, $profile_datafile)
+            or die "Profiling $test failed\n";
 
         if ($opts{html}) {
-            my $cmd = "$perl $nytprofhtml --file=$profile_datafile --out=$outdir";
+            my $htmloutdir = "/tmp/$outdir";
+            unlink <$htmloutdir/*>;
+            my $cmd = "$perl $nytprofhtml --file=$profile_datafile --out=$htmloutdir";
             $cmd .= " --open" if $opts{open};
             run_command($cmd);
         }
     }
+    elsif ($type eq 'rdt') {
+        verify_data($test, $tag, $test_datafile);
+
+        if ($opts{mergerdt}) { # run the file through nytprofmerge
+            my $merged = "$profile_datafile.merged";
+            my $merge_cmd = "$nytprofmerge -v --out=$merged $test_datafile";
+            system($merge_cmd) == 0
+                or die "Error running $merge_cmd\n";
+            verify_data($test, "$tag (merged)", $merged);
+            unlink $merged;
+        }
+    }
+    elsif ($type eq 'x') {
+        mkdir $outdir or die "mkdir($outdir): $!" unless -d $outdir;
+        unlink <$outdir/*>;
+
+        verify_csv_report($test, $tag, $test_datafile, $outdir);
+    }
     elsif ($type =~ /^(?:pl|pm|new|outdir)$/) {
         # skip; handy for "test.pl t/test01.*"
     }
@@ -228,6 +322,7 @@ sub run_test {
     }
 }
 
+
 sub run_command {
     my ($cmd, $show_stdout) = @_;
     warn "NYTPROF=$ENV{NYTPROF}\n" if $opts{v} && $ENV{NYTPROF};
@@ -248,16 +343,23 @@ sub run_command {
 }
 
 
+sub run_perl_command {
+    my ($cmd, $show_stdout) = @_;
+    local $ENV{PERL5LIB} = $perl5lib;
+    run_command("$this_perl $cmd", $show_stdout);
+}
+
+
 sub profile {
     my ($test, $profile_datafile) = @_;
 
     my $cmd = "$perl $opts{profperlopts} $test";
-    ok run_command($cmd), "$test runs ok under the profiler";
+    return ok run_command($cmd), "$test runs ok under the profiler";
 }
 
 
 sub verify_data {
-    my ($test, $profile_datafile) = @_;
+    my ($test, $tag, $profile_datafile) = @_;
 
     my $profile = eval { Devel::NYTProf::Data->new({filename => $profile_datafile}) };
     if ($@) {
@@ -267,14 +369,15 @@ sub verify_data {
     }
 
     SKIP: {
-        skip 'Expected profile data does not have VMS paths',1 if(($^O eq 'VMS') && ($test =~m/test60|test14/i));
-	$profile->normalize_variables;
-        dump_profile_to_file($profile, $test.'_new');
-        my @got      = slurp_file($test.'_new');
-        my @expected = slurp_file($test);
-        is_deeply(\@got, \@expected, "$test match generated profile data")
+        skip 'Expected profile data does not have VMS paths', 1
+            if $^O eq 'VMS' and $test =~ m/test60|test14/i;
+        $profile->normalize_variables;
+        dump_profile_to_file($profile, $test.'_new', $test.'_newp');
+        my @got      = slurp_file($test.'_new'); chomp @got;
+        my @expected = slurp_file($test);        chomp @expected;
+        is_deeply(\@got, \@expected, "$test match generated profile data for $tag")
             ? unlink($test.'_new')
-            : diff_files($test, $test.'_new');
+            : diff_files($test, $test.'_new', $test.'_newp');
     }
 }
 
@@ -290,12 +393,18 @@ sub dump_data_to_file {
 
 
 sub dump_profile_to_file {
-    my ($profile, $file) = @_;
+    my ($profile, $file, $rename_existing) = @_;
+    rename $file, $rename_existing or warn "rename($file, $rename_existing): $!"
+        if $rename_existing && -f $file;
     open my $fh, ">", $file or croak "Can't open $file: $!";
     $profile->dump_profile_data(
         {   filehandle => $fh,
             separator  => "\t",
-            skip_stdlib => 1,
+            skip_fileinfo_hook => sub {
+                my $fi = shift;
+                return 1 if $fi->filename =~ /(AutoLoader|Exporter)\.pm$/ or $fi->filename =~ m!^/\.\.\./!;
+                return 0;
+            },
         }
     );
     return;
@@ -303,16 +412,17 @@ sub dump_profile_to_file {
 
 
 sub diff_files {
+    my ($old_file, $new_file, $newp_file) = @_;
 
     # we don't care if this fails, it's just an aid to debug test failures
     my @opts = split / /, $ENV{NYTPROF_DIFF_OPTS} || '';    # e.g. '-y'
     @opts = ('-u') unless @opts;
-    system("diff @opts @_ 1>&2");
+    system("cmp -s $new_file $newp_file || diff @opts $old_file $new_file 1>&2");
 }
 
 
 sub verify_csv_report {
-    my ($test, $profile_datafile, $outdir) = @_;
+    my ($test, $tag, $profile_datafile, $outdir) = @_;
 
     # generate and parse/check csv report
 
@@ -327,7 +437,7 @@ sub verify_csv_report {
     $csvfile =~ s/\.x//;
     $csvfile .= ".p" unless $csvfile =~ /\.p/;
     $csvfile = html_safe_filename($csvfile);
-    $csvfile = "$outdir/${csvfile}-line.csv";
+    $csvfile = "$outdir/${csvfile}-1-line.csv";
     unlink $csvfile;
 
     my $cmd = "$perl $nytprofcsv --file=$profile_datafile --out=$outdir";
@@ -400,9 +510,11 @@ sub verify_csv_report {
         print "\n";
     }
 
-    is_deeply(\@got, \@expected, "$test match generated CSV data") or do {
-        spit_file($test.'_new', join("", @got));
-        diff_files($test, $test.'_new');
+    chomp @got;
+    chomp @expected;
+    is_deeply(\@got, \@expected, "$test match generated CSV data for $tag") or do {
+        spit_file($test.'_new', join("\n", @got,''), $test.'_newp');
+        diff_files($test, $test.'_new', $test.'_newp');
     };
     is(join("\n", @accuracy_errors), '', "$test times should be reasonable");
 }
@@ -440,7 +552,9 @@ sub slurp_file {    # individual lines in list context, entire file in scalar co
 
 
 sub spit_file {
-    my ($file, $content) = @_;
+    my ($file, $content, $rename_existing) = @_;
+    rename $file, $rename_existing or warn "rename($file, $rename_existing): $!"
+        if $rename_existing && -f $file;
     open my $fh, ">", $file or croak "Can't open $file: $!";
     print $fh $content;
     close $fh or die "Error closing $file: $!";
@@ -472,6 +586,7 @@ sub count_of_failed_tests {
 }
 
 
+
 1;
 
-# vim:ts=8:sw=4
+# vim:ts=8:sw=4:et
